#ifndef aegis256x2_neon_aes_H
#define aegis256x2_neon_aes_H

#include "../common/common.h"
#include "implementations.h"

extern struct aegis256x2_implementation aegis256x2_neon_aes_implementation;

#endif
