#ifndef aegis128x4_altivec_H
#define aegis128x4_altivec_H

#include "../common/common.h"
#include "implementations.h"

extern struct aegis128x4_implementation aegis128x4_altivec_implementation;

#endif
