#ifndef aegis128l_neon_aes_H
#define aegis128l_neon_aes_H

#include "../common/common.h"
#include "implementations.h"

extern struct aegis128l_implementation aegis128l_neon_aes_implementation;

#endif
