#ifndef aegis256x2_altivec_H
#define aegis256x2_altivec_H

#include "../common/common.h"
#include "implementations.h"

extern struct aegis256x2_implementation aegis256x2_altivec_implementation;

#endif
